/* ----------------------------------------------------------------------------
 *
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 *
 * ----------------------------------------------------------------------------
 */

#include <stdio.h>

/* rdm.h is required in all files using the RDM Database */
#include "rdm.h"

/* core13_db.h was generated from the DDL file core13_db.ddl by the ddlp.
 * utility.
 */
#include "core13_db.h"

/* core13_db_dbd.h was generated from the DDL
 * file core13_db.ddl by the ddlp utility.  It is required to use the d_open_ptr API
 */
#include "core13_db_dbd.h"

/* EXAMPLE - Core for C */
/* Each time you run this example:
 *   1) The database is initialized (all existing contents removed)
 *   2) Two RDM tasks are created
 *   3) Each task opens the database is shared mode
 *   4) Task 1 starts a transaction, locks the office record for writes,
 *      adds a few office records
 *   5) Task 2 attempts to read the office records (fails while task 1 has lock)
 *   6) Task 1 commits transaction
 *   7) Task 2 is able to read the office records
 *
 * The DDL (Database Definition Language) is in core13_db.ddl. The database
 * schema is embedded into the application and the database will be created 
 * the first time the application is run.  Subsequent runs will use the now
 * existing database
 *
 * The following RDM Functions are illustrated:
 *      d_opentask/d_closetask
 *      d_open_ptr/d_close
 *      d_initialize
 *      d_reclock/d_recfree
 *      d_trbegin/d_trend
 *      d_fillnew
 *      d_keyfrst/d_keynext
 *      d_recread
 *      d_errorname
 *      d_errorstr
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 *
 * Each of these functions returns an integer status code, where the
 * value S_OKAY indicates a successful call.
 *
 * For simplicity, this example does not check all return codes, but good
 * programming practices would dictate that they are checked after each
 * RDM call.
 *
 * The actual database will be stored in a directory named 'core13_db'
 */

/* This function displays the name of each OFFICE record instance.
*/
int32_t display_offices(
    DB_TASK *task,
    int32_t  dbn)
{
    int32_t       rdm_status;
    struct office office_rec;

    /* Get a read-lock on the office record */    
    rdm_status = d_reclock(OFFICE, "r", task, CURR_DB);
    if ( rdm_status != S_OKAY)
    {
        fprintf (stderr, "Error read locking OFFICE records - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status) );
        /* We should return the here, but we will try to read anyway to see 
         * what happens if we read without the lock
         */
    }

    for (rdm_status = d_keyfrst (OFFICE_NAME, task, dbn);
         rdm_status == S_OKAY;
         rdm_status = d_keynext (OFFICE_NAME, task, dbn))
    {
        /* Read and display the current person record */
        d_recread(&office_rec, task, dbn);
        printf("%s\n", office_rec.name);
    }

    /* Expect rdm_status to be S_NOTFOUND when we exit the loop */
    if(rdm_status == S_NOTFOUND)
    {
        rdm_status = S_OKAY;
    }
    else
    {
        fprintf (stderr, "Error reading OFFICE records - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status) );
    }

    /* Free the read lock */
    d_recfree(OFFICE, task, dbn);

    return rdm_status;
}

/* =======================================================================
*/
int32_t main(void)
{
    DB_TASK       *task1;
    DB_TASK       *task2;
    int32_t        rdm_status;
    struct office  office_rec;

    /* create the contexts for the database operations */
    d_opentask(&task1);
    d_opentask(&task2);

    /* Open the database for shared access, this will create the database
     * dictionary file (core13_db.dbd) if it does not already exist.
     */
    rdm_status = d_open_ptr("core13_db", "s", &core13_db_dbd, sizeof(core13_db_dbd), task1);
    if (rdm_status != S_OKAY)
    {
        fprintf (stderr, "Shared open failed for task1, error = %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status) );
        d_closetask (task1);
        d_closetask (task2);
        return (1);
    }

    /* Open the database for shared access, the database dictionary file
     * (core13_db.dbd) should now exist.
     */
    rdm_status = d_open_ptr("core13_db", "s", &core13_db_dbd, sizeof(core13_db_dbd), task2);
    if (rdm_status != S_OKAY)
    {
        fprintf (stderr, "Shared open failed for task1, error = %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status) );
        d_closetask (task1);
        d_closetask (task2);
        return (1);
    }

    /* Initialize the databases.  Initializing will create the data files if 
     * they do not exist.  If the the data files do already exist they will be
     * removed and new empty files will be created.  Since we are opened in 
     * shared mode we need to use a transaction and locks
     */
    d_trbegin("InitalizeDB", task1);
    d_reclock(OFFICE, "w", task1, CURR_DB);
    d_initialize(task1, CURR_DB);
    d_trend(task1);

    /* Start a database transaction in task 1*/
    rdm_status = d_trbegin("Add NA Offices", task1);
    if (rdm_status != S_OKAY)
    {
        fprintf (stderr, "Error starting a transaction, error = %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status) );
        d_close(task1);
        d_close(task2);
        d_closetask (task1);
        d_closetask (task2);
        return (1);
    }

    /* Get a write lock on the OFFICE record */
    d_reclock(OFFICE, "w", task1, CURR_DB);

    /* Add NA offices to the database */
    strncpy (office_rec.name, "Seattle", SIZEOF_OFFICE_NAME);
    rdm_status = d_fillnew (OFFICE, &office_rec, task1, CURR_DB);

    strncpy (office_rec.name, "Boise", SIZEOF_OFFICE_NAME);
    rdm_status = d_fillnew (OFFICE, &office_rec, task1, CURR_DB);

    strncpy (office_rec.name, "San Francisco", SIZEOF_OFFICE_NAME);
    rdm_status = d_fillnew (OFFICE, &office_rec, task1, CURR_DB);

    strncpy (office_rec.name, "Dallas", SIZEOF_OFFICE_NAME);
    rdm_status = d_fillnew (OFFICE, &office_rec, task1, CURR_DB);
    
    /* Display all offices using task2 (should fail because we will not be 
     * able to get a read lock */
    printf("\nDisplaying Offices using task2.\n");
    display_offices(task2, CURR_DB);
   
    /* Commit the database transaction in task1 */
    rdm_status = d_trend(task1);
    if (rdm_status != S_OKAY)
    {
        fprintf (stderr, "Error committing a transaction, error = %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status) );
        d_close(task1);
        d_close(task2);
        d_closetask(task1);
        d_closetask(task2);

        return (1);
    }
        
    /* Display all offices using task 2 (should succeed because task1 has
     * freed write locks by ending the transaction
     */
    printf("\nDisplaying Offices using task2.\n");
    display_offices(task2, CURR_DB);

    /* close the database */
    d_close(task1);
    d_close(task2);
    d_closetask(task1);
    d_closetask(task2);
    
    printf("\nPress <Enter> to continue . . .\n");
    getchar();

    return EXIT_SUCCESS;
}
